"""Cache-Controlled Z SQL Methods

Z SQL Methods that share a cache across threads of a Zope server,
allowing the cache to be cleared for all threads.


@License: GPL
@Author: Dieter Maurer <dieter@handshake.de>
@Author: Joel Burton <joel@joelburton.com>
@Contact: Joel Burton <joel@joelburton.com>
"""

# $Id: $

import Shared.DC.ZRDB.DA

from Products.ZSQLMethods.SQL import SQLConnectionIDs
import Products.ZSQLMethods.SQL
from Globals import HTMLFile
import Globals
from SharedCache import getResource
from new import function
from Shared.DC.ZRDB.Aqueduct import parse
from IOBTree import Bucket
from time import time


manage_addCCSQLMethodForm=HTMLFile('dtml/addCCSQL', globals())
manage_addCCSQLMethodForm.__name__='addCCSQL'

def manage_addCCSQLMethod(self, id, title,
                                connection_id, arguments, template,
                                REQUEST=None, submit=None):
    """Add an SQL Method

    @type  connection_id: string
    @param connection_id: the id of a database connection
        that resides in the current folder or in a folder above the
        current folder.  The database should understand SQL.

    @type  arguments: string
    @param arguments: arguments specifications, as would be given
        in the SQL method cration form.

    @type  template: string
    @param template: source for the SQL template.
    """

    self._setObject(id, SQL(id, title, connection_id, arguments, template))
    if REQUEST is not None:
        u=REQUEST['URL1']
        if submit=="Add and Edit":
            u="%s/%s/manage_main" % (u,id)
        elif submit=="Add and Test":
            u="%s/%s/manage_testForm" % (u,id)
        else:
            u=u+'/manage_main'
            
        REQUEST.RESPONSE.redirect(u)
    return ''

class SQL(Shared.DC.ZRDB.DA.DA):
    """Cache-controlled ZSQL Method.

    Z SQL Methods that share a cache across threads of a Zope server,
    allowing the cache to be cleared for all threads.
    """
    meta_type='CC SQL Method'
                
    manage_main=Products.ZSQLMethods.SQL.SQL.manage_main

    __ac_permissions__=(
      ('Use Database Methods',
       ('flushCache', 'flushCacheEntry'),
       ('Anonymous','Manager')),
      )

    def manage_edit(self,title,connection_id,arguments,template,
                    SUBMIT='Change',sql_pref__cols='50', sql_pref__rows='20',
                    REQUEST=None):
        """Change database method  properties

        @type  connection_id: string
        @param connection_id: the id of a database connection
            that resides in the current folder or in a folder above the
            current folder.  The database should understand SQL.

        @type  arguments: string
        @param arguments: arguments specifications, as would be given
            in the SQL method cration form.

        @type  template: string
        @param template: source for the SQL template.
        """

        if self._size_changes.has_key(SUBMIT):
            return self._er(title,connection_id,arguments,template,
                            SUBMIT,sql_pref__cols,sql_pref__rows,REQUEST)

        self.title=title
        self.connection_id=connection_id
        self.arguments_src=arguments
        self._arg=parse(arguments)
        self.src=template
        self.template=t=self.template_class(template)
        t.cook()

        # DM
        #self._v_cache={}, Bucket()
        # do not do it here, 'self' may not yet be acquisition wrapped!
        self._flushCache=self._flushCache+1

        if REQUEST:
            if SUBMIT=='Change and Test':
                return self.manage_test(REQUEST)
            message='ZSQL Method content changed'
            return self.manage_main(self, REQUEST, manage_tabs_message=message)
        return ''

    def _cached_result(self, DB__, query):

        # Try to fetch from cache
        cache, tcache = self.getCache()
        max_cache=self.max_cache_
        now=time()
        t=now-self.cache_time_
        if len(cache) > max_cache / 2:
            keys=tcache.keys()
            keys.reverse()
            while keys and (len(keys) > max_cache or keys[-1] < t):
                key=keys[-1]
                # DM: because we now use shared caches, the operations
                #     below may not succeed
                try:
                    q=tcache[key]
                    try: del tcache[key]
                    except KeyError: pass
                    try: del cache[q]
                    except KeyError: pass
                except KeyError: pass
                del keys[-1]
                
        # DM: 'if cache.has_key' replaced by 'try' to avoid race conditions.
        try:
            k, r = cache[query]
            if k > t: return r
        except KeyError: pass

        result=apply(DB__.query, query)
        if self.cache_time_ > 0:
            tcache[int(now)]=query
            cache[query]= now, result

        return result

    def getCache(self):
        """Public wrapper for cache.  
        
        Checks to see if it needs to be flushed.
        """

        #if self._flushCache: self.flushCache()
        if self._flushCache > getattr(self, '_v_flushCacheCount' ,0):
            self._doFlushCache()
            self._v_flushCacheCount=self._flushCache
        return self._getCache()

    def _getCache(self):
        if hasattr(self,'_v_cache'): 
            cache=self._v_cache
        else:
            # DM: use shared cache, bucket
            physPath= _getPhysicalPath(self)
            cache=self._v_cache= (\
              getResource(physPath + ('_cache',),lambda : {}),
              getResource(physPath + ('_bucket',), Bucket)
              )
        return cache
      
    _flushCache= 0
    def _doFlushCache(self):
        """Actually perform the flushing."""
        
        cache, tcache=self._getCache()
        cache.clear(); tcache.clear()
        #self._flushCache= 0

    def flushCache(self):
        """Request that the cache be flushed.

        Mark our internal _flushCache attr and perform the flush.
        Other ZEO clients, if any, will see the marked-up ._flushCache
        and flush their cache when they first try to use it.
        """

        self._flushCache = self._flushCache + 1
        self._doFlushCache()
        self._v_flushCacheCount = self._flushCache

    def flushCacheEntry(self, REQUEST=None, **kw):
        """Flush just this entry from the cache.

        :Parameters:

          - `**kw`: Same parameters you pass to the SQLMethod

          - `REQUEST`: Current REQUEST.  Needed in case SQLMethod looks
            in REQUEST to render itself.

        :Warn: 
          This does not work across ZEO clients.  Only the client
          that this is called in will flush this cache entry. You have
          been warned.
        """

        kw=kw.copy()
        kw['src__']= 1
        query=apply(self.__call__, (REQUEST,), kw)
        cache, _= self.getCache()
        try:
            del cache[(query,self.max_rows_)]
        except KeyError: 
            pass


Globals.default__class_init__(SQL)



def _getPhysicalPath(object):
    o= object
    gp= getattr(o,'getPhysicalPath',None)
    if gp is not None: 
        return gp()
    from sys import stdout
    l= []
    while o:
        try: id= o.id
        except AttributeError: break
        if callable(id): id= id()
        l.append(id)
        o= getattr(o,'aq_inner',None)
        if o is not None: o= o.aq_parent
    l.reverse()
    return tuple(l)
